/**
 * Models for JsonSchema with derivations in Draft-07, 2019-09, and 2020-12
 *
 * @since 1.2.0
 * @see https://json-schema.org/draft/2020-12/json-schema-validation.html
 */
import { type Const } from 'fp-ts/Const';
import type * as RR from 'fp-ts/ReadonlyRecord';
import { type Float, type MaxNegativeFloat, type MaxPositiveFloat } from 'schemata-ts/float';
import { type Integer, type MaxSafeInt, type MinSafeInt } from 'schemata-ts/integer';
import * as I from 'schemata-ts/internal/json-schema';
import { type Schema } from 'schemata-ts/Schema';
/**
 * @since 1.2.0
 * @category Model
 */
export type JsonSchemaValue = I.JsonEmpty | I.JsonString | I.JsonNumber | I.JsonBoolean | I.JsonNull | I.JsonInteger | I.JsonConst | I.JsonLiteral | I.JsonStruct | I.JsonArray | I.JsonUnion | I.JsonIntersection | I.JsonRef;
/**
 * @since 1.2.0
 * @category Model
 */
export type JsonSchema = JsonSchemaValue & I.Description & I.References & I.Default;
/**
 * Interprets a schema as a JsonSchema projecting into either Draft-07 or 2020-12
 *
 * @since 1.2.0
 * @category Interpreters
 */
export declare const deriveJsonSchema: <I, O>(schema: Schema<I, O>, version?: 'Draft-07' | '2019-09' | '2020-12', maintainIdentity?: boolean) => JsonSchema;
/**
 * @since 1.2.0
 * @category Constructors
 */
export declare const emptySchema: Const<I.JsonEmpty, never>;
/**
 * @since 1.2.0
 * @category Constructors
 */
export declare const string: (params?: {
    minLength?: number;
    maxLength?: number;
    pattern?: string;
    contentEncoding?: string;
    contentMediaType?: string;
    contentSchema?: JsonSchema;
    format?: string;
}) => Const<JsonSchema, string>;
/**
 * @since 1.2.0
 * @category Constructors
 */
export declare const number: <Min extends number | undefined = undefined, Max extends number | undefined = undefined>(params?: {
    minimum?: Min | undefined;
    maximum?: Max | undefined;
}) => Const<JsonSchema, Float<Min extends undefined ? -1.7976931348623157e+308 : Min, Max extends undefined ? 1.7976931348623157e+308 : Max>>;
/**
 * @since 1.2.0
 * @category Constructors
 */
export declare const integer: <Min extends number | undefined = undefined, Max extends number | undefined = undefined>(params?: {
    minimum?: Min | undefined;
    maximum?: Max | undefined;
}) => Const<JsonSchema, Integer<Min extends undefined ? -9007199254740991 : Min, Max extends undefined ? 9007199254740991 : Max>>;
/**
 * @since 1.2.0
 * @category Constructors
 */
export declare const booleanSchema: Const<JsonSchema, boolean>;
/**
 * @since 1.2.0
 * @category Constructors
 */
export declare const literal: <A extends string | number | boolean | null>(value: A) => Const<JsonSchema, A>;
/**
 * @since 1.2.0
 * @category Constructors
 */
export declare const struct: <A>(properties: { [K in keyof A]: Const<JsonSchema, A[K]>; }, required?: ReadonlyArray<string>, additionalProperties?: JsonSchema | false) => Const<JsonSchema, A>;
/**
 * @since 1.2.0
 * @category Constructors
 */
export declare const record: <A>(additionalProperties: Const<JsonSchema, A>, propertyNames?: Const<JsonSchema, string>) => Const<JsonSchema, Record<string, A>>;
/**
 * @since 1.2.0
 * @category Constructors
 */
export declare const array: (params?: {
    minItems?: number;
    maxItems?: number;
}) => <A>(items: Const<JsonSchema, A>) => Const<JsonSchema, A[]>;
/**
 * @since 1.2.0
 * @category Constructors
 */
export declare const tuple: <A extends readonly unknown[]>(...items: { [K in keyof A]: Const<JsonSchema, A[K]>; }) => Const<JsonSchema, A>;
/**
 * @since 1.2.0
 * @category Constructors
 */
export declare const nullSchema: Const<I.JsonNull, never>;
/**
 * @since 1.2.0
 * @category Constructors
 */
export declare const union: <U extends readonly Const<JsonSchema, unknown>[]>(...members: U) => Const<JsonSchema, U[number] extends Const<JsonSchema, infer A> ? A : never>;
/**
 * @since 1.2.0
 * @category Constructors
 */
export declare const intersection: <B>(right: Const<JsonSchema, B>) => <A>(left: Const<JsonSchema, A>) => Const<JsonSchema, A & B>;
/**
 * A reference to a schema definition
 *
 * @since 2.0.0
 * @category Constructors
 */
export declare const ref: <A>(ref: string) => Const<JsonSchema, A>;
/**
 * @since 1.2.0
 * @category URI
 */
export declare const URI = "schemata-ts/JsonSchema";
/**
 * @since 1.2.0
 * @category URI
 */
export type URI = typeof URI;
declare module 'fp-ts/lib/HKT' {
    interface URItoKind2<E, A> {
        readonly JsonSchema: Const<JsonSchema, E>;
    }
}
/**
 * @since 1.2.0
 * @category Combintators
 */
export declare const annotate: (params?: {
    readonly title?: string;
    readonly description?: string;
    readonly references?: RR.ReadonlyRecord<string, JsonSchema>;
    readonly deprecated?: boolean;
    readonly readOnly?: boolean;
}) => (schema: JsonSchema) => Const<JsonSchema, never>;
/**
 * Removes the internal class identities from a `JsonSchema`
 *
 * @since 1.2.0
 * @category Destructors
 */
export declare const stripIdentity: (schema: JsonSchema) => JsonSchema;
//# sourceMappingURL=JsonSchema.d.ts.map