/** @since 2.2.0 */
import * as E from "fp-ts/es6/Either.js";
import { deriveGuard, deriveInputGuard } from 'schemata-ts/Guard';
import { drawTree, safeShow } from 'schemata-ts/TranscodeError';
import { deriveTranscoder } from 'schemata-ts/Transcoder';
import { deriveTypeString } from 'schemata-ts/TypeString';
/**
 * A trait which will throw an error for when an expected "output" value does not match a
 * schema's output type
 *
 * @since 2.2.0
 * @category Interpreters
 */

export const deriveAssert = schema => {
  const guard = deriveGuard(schema);
  const [, name] = deriveTypeString(schema);
  return {
    assert: output => {
      if (!guard.is(output)) {
        throw new TypeError(`Expected ${name}, but got ${safeShow(output)}`);
      }
    }
  };
};
/**
 * A trait which will throw an error for when an expected "input" value does not match a
 * schema's input type
 *
 * @since 2.2.0
 * @category Interpreters
 */

export const deriveInputAssert = schema => {
  const guard = deriveInputGuard(schema);
  const [name] = deriveTypeString(schema);
  return {
    assert: input => {
      if (!guard.is(input)) {
        throw new TypeError(`Expected ${name}, but got ${safeShow(input)}`);
      }
    }
  };
};
/**
 * A trait which will throw an error for when an expected "input" value does not match a
 * schema's input type
 *
 * Unlike `deriveInputAssert`, this will include the error tree in the error message
 *
 * @since 2.2.0
 * @category Interpreters
 */

export const deriveInputAssertTree = schema => {
  const transcoder = deriveTranscoder(schema);
  const [name] = deriveTypeString(schema);
  return {
    assert: input => {
      const result = transcoder.decode(input);

      if (E.isLeft(result)) {
        throw new TypeError(`Expected ${name}, but received the following errors:\n${drawTree(result.left, {
          showHeading: false
        })}`);
      }
    }
  };
};
//# sourceMappingURL=Assert.mjs.map