/**
 * Models for JsonSchema with derivations in Draft-07, 2019-09, and 2020-12
 *
 * @since 1.2.0
 * @see https://json-schema.org/draft/2020-12/json-schema-validation.html
 */
import { make } from "fp-ts/es6/Const.js";
import { identity } from "fp-ts/es6/function.js";
import { deriveJsonSchema as deriveJsonSchema_ } from 'schemata-ts/derivations/json-schema-schemable';
import * as I from 'schemata-ts/internal/json-schema'; // -------------------------------------------------------------------------------------
// Constructors
// -------------------------------------------------------------------------------------

/**
 * Interprets a schema as a JsonSchema projecting into either Draft-07 or 2020-12
 *
 * @since 1.2.0
 * @category Interpreters
 */

export const deriveJsonSchema = (schema, version = '2019-09', maintainIdentity = false) => {
  switch (version) {
    case 'Draft-07':
      // istanbul ignore next
      return (maintainIdentity ? identity : I.stripIdentity)(I.asDraft07(deriveJsonSchema_(schema)));

    case '2019-09':
      // istanbul ignore next
      return (maintainIdentity ? identity : I.stripIdentity)(deriveJsonSchema_(schema));

    case '2020-12':
      // istanbul ignore next
      return (maintainIdentity ? identity : I.stripIdentity)(I.as2020(deriveJsonSchema_(schema)));
  }
};
/**
 * @since 1.2.0
 * @category Constructors
 */

export const emptySchema = /*#__PURE__*/make( /*#__PURE__*/new I.JsonEmpty());
/**
 * @since 1.2.0
 * @category Constructors
 */

export const string = (params = {}) => make(new I.JsonString(params.minLength, params.maxLength, params.pattern, params.contentEncoding, params.contentMediaType, params.contentSchema, params.format));
/**
 * @since 1.2.0
 * @category Constructors
 */

export const number = (params = {}) => make(new I.JsonNumber(params.minimum, params.maximum));
/**
 * @since 1.2.0
 * @category Constructors
 */

export const integer = (params = {}) => make(new I.JsonInteger(params.minimum, params.maximum));
/**
 * @since 1.2.0
 * @category Constructors
 */

export const booleanSchema = /*#__PURE__*/make( /*#__PURE__*/new I.JsonBoolean());
/**
 * @since 1.2.0
 * @category Constructors
 */

export const literal = value => value === null ? make(new I.JsonNull()) : make({
  type: typeof value,
  const: value
});
/**
 * @since 1.2.0
 * @category Constructors
 */

export const struct = (properties, required = [], additionalProperties) => make(new I.JsonStruct(properties, required, additionalProperties));
/**
 * @since 1.2.0
 * @category Constructors
 */

export const record = (additionalProperties, propertyNames) => make(new I.JsonStruct(new I.JsonEmpty(), [], additionalProperties, propertyNames));
/**
 * @since 1.2.0
 * @category Constructors
 */

export const array = (params = {}) => items => make(new I.JsonArray(items, params.minItems, params.maxItems));
/**
 * @since 1.2.0
 * @category Constructors
 */

export const tuple = (...items) => make(new I.JsonArray(items, items.length, items.length));
/**
 * @since 1.2.0
 * @category Constructors
 */

export const nullSchema = /*#__PURE__*/make( /*#__PURE__*/new I.JsonNull());
/**
 * @since 1.2.0
 * @category Constructors
 */

export const union = (...members) => make(members.length > 1 ? new I.JsonUnion(members) : members[0]);
/**
 * @since 1.2.0
 * @category Constructors
 */

export const intersection = right => left => make(new I.JsonIntersection([left, right]));
/**
 * A reference to a schema definition
 *
 * @since 2.0.0
 * @category Constructors
 */

export const ref = ref => make(new I.JsonRef(ref)); // -------------------------------------------------------------------------------------
// Instances
// -------------------------------------------------------------------------------------

/**
 * @since 1.2.0
 * @category URI
 */

export const URI = 'schemata-ts/JsonSchema'; // -------------------------------------------------------------------------------------
// Combintators
// -------------------------------------------------------------------------------------

/**
 * @since 1.2.0
 * @category Combintators
 */
// istanbul ignore next

export const annotate = ({
  title,
  description,
  references,
  deprecated,
  readOnly
} = {}) => schema => title === undefined && description === undefined && references === undefined && deprecated === undefined && readOnly === undefined ? make(schema) : make({ ...schema,
  ...(title === undefined ? {} : {
    title
  }),
  ...(description === undefined ? {} : {
    description
  }),
  ...(references === undefined ? {} : {
    $defs: references
  }),
  ...(deprecated === undefined ? {} : {
    deprecated
  }),
  ...(readOnly === undefined ? {} : {
    readOnly
  })
}); // -------------------------------------------------------------------------------------
// Destructors
// -------------------------------------------------------------------------------------

/**
 * Removes the internal class identities from a `JsonSchema`
 *
 * @since 1.2.0
 * @category Destructors
 */

export const stripIdentity = I.stripIdentity;
//# sourceMappingURL=JsonSchema.mjs.map