/**
 * Parses a string and turns it into camelCase.
 *
 * @since 1.4.0
 */
import { tailRec } from "fp-ts/es6/ChainRec.js";
import * as E from "fp-ts/es6/Either.js";
import { flow, identity, pipe } from "fp-ts/es6/function.js";
import * as O from "fp-ts/es6/Option.js";
import * as Pred from "fp-ts/es6/Predicate.js";
import * as RA from "fp-ts/es6/ReadonlyArray.js";
import * as Str from "fp-ts/es6/string.js";
import * as k from 'kuvio';
/** Convert a character code to character */

const cc = String.fromCharCode;
/** A list of whitespace + "-" + "_" patterns to match type-fest's type-level camelcase */

const delimiters = /*#__PURE__*/k.characterClass(false, /*#__PURE__*/cc(9),
/*#__PURE__*/
// tab
cc(10),
/*#__PURE__*/
// line feed
cc(11),
/*#__PURE__*/
// vertical tab
cc(12),
/*#__PURE__*/
// form feed
cc(13),
/*#__PURE__*/
// carriage return
cc(32),
/*#__PURE__*/
// space
cc(133),
/*#__PURE__*/
// next line
cc(160),
/*#__PURE__*/
// non-breaking space
cc(5760),
/*#__PURE__*/
// ogham space mark
cc(8192),
/*#__PURE__*/
// en quad
cc(8193),
/*#__PURE__*/
// em quad
cc(8194),
/*#__PURE__*/
// en space
cc(8195),
/*#__PURE__*/
// em space
cc(8196),
/*#__PURE__*/
// three-per-em space
cc(8197),
/*#__PURE__*/
// four-per-em space
cc(8198),
/*#__PURE__*/
// six-per-em space
cc(8199),
/*#__PURE__*/
// figure space
cc(8200),
/*#__PURE__*/
// punctuation space
cc(8201),
/*#__PURE__*/
// thin space
cc(8202),
/*#__PURE__*/
// hair space
cc(8232),
/*#__PURE__*/
// line separator
cc(8233),
/*#__PURE__*/
// paragraph separator
cc(8239),
/*#__PURE__*/
// narrow no-break space
cc(8287),
/*#__PURE__*/
// medium mathematical space
cc(12288),
/*#__PURE__*/
// ideographic space
cc(65279), // zero width no-break space
'-', // hyphen-minus
'_');
const delimiterRegex = /*#__PURE__*/k.regexFromPattern(delimiters);

const dropRightWhile = predicate => as => pipe(as, RA.findLastIndex(Pred.not(predicate)), O.fold(() => [], i => as.slice(0, i + 1)));

const trimDelimiters = /*#__PURE__*/flow(Str.trim, /*#__PURE__*/Str.split(''), /*#__PURE__*/RA.dropLeftWhile(s => delimiterRegex.test(s)), /*#__PURE__*/dropRightWhile(s => delimiterRegex.test(s)), /*#__PURE__*/RA.foldMap(Str.Monoid)(identity));

const isLetter = s => /[a-zA-Z]/.test(s);

const isUppercase = s => isLetter(s) && s === s.toUpperCase();

const isNumber = char => /[0-9]/.test(char);

const hasLowercase = s => /[a-z]/gm.test(s);
/**
 * Parses a string and turns it into camelCase.
 *
 * @since 1.4.0
 */


export const camelCase = input => {
  if (input.length === 0) return input;

  const go = s => {
    const char = s.input.charAt(s.index);
    const nextChar = pipe(s.input.charAt(s.index + 1), O.fromPredicate(s => s.length > 0)); // Stream is exhausted

    if (char === '') {
      return E.right(s.output);
    } // Lowercase the first char


    if (s.index === 0) {
      return E.left({
        input: s.input,
        output: s.output + char.toLowerCase(),
        index: s.index + 1,
        lastCharWasDelimiter: false,
        lastCharWasNumber: isNumber(char),
        hasBeenScreaming: isUppercase(char)
      });
    } // If the character is a delimiter, skip it


    if (delimiterRegex.test(char)) {
      return E.left({
        input: s.input,
        output: s.output,
        index: s.index + 1,
        lastCharWasDelimiter: true,
        lastCharWasNumber: false,
        hasBeenScreaming: false
      });
    } // If the last char was screaming, lowercase the current char if the next char is uppercase


    if (s.hasBeenScreaming && isUppercase(char)) {
      return E.left({
        input: s.input,
        output: s.output + (O.isSome(nextChar) && isUppercase(nextChar.value) ? char.toLowerCase() : char),
        index: s.index + 1,
        lastCharWasDelimiter: false,
        lastCharWasNumber: isNumber(char),
        hasBeenScreaming: true
      });
    } // If the last char was a delimiter (or number), uppercase the current char


    if (s.lastCharWasDelimiter || s.lastCharWasNumber) {
      return E.left({
        input: s.input,
        output: s.output + char.toUpperCase(),
        index: s.index + 1,
        hasBeenScreaming: isUppercase(char),
        lastCharWasNumber: isNumber(char),
        lastCharWasDelimiter: false
      });
    }

    return E.left({
      input: s.input,
      output: s.output + char,
      index: s.index + 1,
      lastCharWasDelimiter: false,
      lastCharWasNumber: isNumber(char),
      hasBeenScreaming: isUppercase(char)
    });
  };

  const trimmed = trimDelimiters(input);
  if (trimmed.length === 0) return trimmed;
  return tailRec({
    input: hasLowercase(trimmed) ? trimmed : trimmed.toLowerCase(),
    output: '',
    index: 0,
    lastCharWasDelimiter: false,
    lastCharWasNumber: false,
    hasBeenScreaming: isUppercase(trimmed.charAt(0))
  }, go);
};
//# sourceMappingURL=camelcase.mjs.map