import { make as make_ } from "fp-ts/es6/Const.js";
import * as E from "fp-ts/es6/Either.js";
import { flow, pipe, tuple } from "fp-ts/es6/function.js";
import * as RA from "fp-ts/es6/ReadonlyArray.js";
import * as RNEA from "fp-ts/es6/ReadonlyNonEmptyArray.js";
import * as RR from "fp-ts/es6/ReadonlyRecord.js";
import { hasOwn } from 'schemata-ts/internal/util';
export const make = make_;
/** @internal */

export const addDescription = description => schema => make_(Object.assign({}, schema, description));
/** @internal */

export const addReferences = references => schema => make_(Object.assign({}, schema, references));
/** @internal */

export const addDefault = defaultValue => schema => make_(Object.assign({}, schema, {
  default: defaultValue
}));
/** Matches anything */

export class JsonEmpty {}
/** Matches a subset of strings */

export class JsonString {
  minLength;
  maxLength;
  pattern;
  contentEncoding;
  contentMediaType;
  contentSchema;
  format;
  type = 'string';

  constructor(minLength, maxLength, pattern, contentEncoding, contentMediaType, contentSchema, format) {
    this.minLength = minLength;
    this.maxLength = maxLength;
    this.pattern = pattern;
    this.contentEncoding = contentEncoding;
    this.contentMediaType = contentMediaType;
    this.contentSchema = contentSchema;
    this.format = format;
  }

}
/** Matches a subset of floats */

export class JsonNumber {
  minimum;
  maximum;
  type = 'number';

  constructor(minimum, maximum) {
    this.minimum = minimum;
    this.maximum = maximum;
  }

}
/** Matches a subset of integers */

export class JsonInteger {
  minimum;
  maximum;
  type = 'integer';

  constructor(minimum, maximum) {
    this.minimum = minimum;
    this.maximum = maximum;
  }

}
/** Matches true or false */

export class JsonBoolean {
  type = 'boolean';
}
/** Matches a constant value */

export class JsonConst {
  const;
}
/** Matches a set of properties with a given set of required properties. */

export class JsonStruct {
  properties;
  required;
  additionalProperties;
  propertyNames;
  type = 'object';

  constructor(properties, required, additionalProperties, propertyNames) {
    this.properties = properties;
    this.required = required;
    this.additionalProperties = additionalProperties;
    this.propertyNames = propertyNames;
  }

}
/** Matches a subset of arrays with uniform index values (or specific index values) */

export class JsonArray {
  items;
  minItems;
  maxItems;
  type = 'array';

  constructor(items, minItems, maxItems) {
    this.items = items;
    this.minItems = minItems;
    this.maxItems = maxItems;
  }

}
/** Matches null exactly */

export class JsonNull {
  type = 'null';
  const = null;
}
/** Matches any of the supplied schemas */

export class JsonUnion {
  anyOf;

  constructor(anyOf) {
    this.anyOf = anyOf;
  }

}
/** Matches all of the supplied schemas */

export class JsonIntersection {
  allOf;

  constructor(allOf) {
    this.allOf = allOf;
  }

}
/**
 * A reference to a named schema definition
 *
 * @since 2.0.0
 */

export class JsonRef {
  $ref;

  constructor($ref) {
    this.$ref = $ref;
  }

}
/** @internal */

export const stripIdentity = schema => JSON.parse(JSON.stringify(schema));
/** @internal */
// istanbul ignore next

class References_ {
  $defs;

  constructor($defs) {
    this.$defs = $defs;
  }

}
/** A recursive key setter */


const recurseKeyOf = (_, key, mapper) => {
  return method => schema => {
    if (!hasOwn(schema, key) || schema[key] === undefined) {
      return schema;
    }

    const {
      [key]: value,
      ...rest
    } = schema;
    const result = mapper(method)(value);

    if (E.isRight(result)) {
      const [newKey, _, inject] = result.right;
      return Object.assign({}, rest, {
        [newKey]: _
      }, inject);
    }

    const [_, inject] = result.left;
    return Object.assign({}, rest, {
      [key]: _
    }, inject);
  };
};

const foldDefault = method => schema => method(schema);

const foldDefaultLeft = method => flow(foldDefault(method), tuple, E.left); // -----------------
// Json Remappers
// -----------------


const remapContentSchema = /*#__PURE__*/recurseKeyOf(JsonString, 'contentSchema', foldDefaultLeft);
const remapProperties = /*#__PURE__*/recurseKeyOf(JsonStruct, 'properties', method => flow(RR.map(method), tuple, E.left));
const remapAdditionalProperties = /*#__PURE__*/recurseKeyOf(JsonStruct, 'additionalProperties', foldDefaultLeft);
const remapPropertyNames = /*#__PURE__*/recurseKeyOf(JsonStruct, 'propertyNames', foldDefaultLeft);
const remapItems = /*#__PURE__*/recurseKeyOf(JsonArray, 'items', method => items => E.left(tuple(Array.isArray(items) ? items.map(method) : method(items))));
const remapAnyOf = /*#__PURE__*/recurseKeyOf(JsonUnion, 'anyOf', _ => flow(RA.map(_), tuple, E.left));
const remapAllOf = /*#__PURE__*/recurseKeyOf(JsonIntersection, 'allOf', _ => flow(RNEA.map(_), tuple, E.left));
const remapDefs = /*#__PURE__*/recurseKeyOf(References_, '$defs', _ => flow(RR.map(_), tuple, E.left)); // ---------------------
// Json Schema Draft-07
// ---------------------

/** @internal */

export const asDraft07 = schema => pipe(schema, defsToDefinitions(asDraft07), recurseDraft07);
const defsToDefinitions = /*#__PURE__*/recurseKeyOf(References_, '$defs', () => flow(RR.map(asDraft07), _ => E.right(tuple('definitions', _))));
const recurseDraft07 = /*#__PURE__*/flow( /*#__PURE__*/remapContentSchema(asDraft07), /*#__PURE__*/remapProperties(asDraft07), /*#__PURE__*/remapAdditionalProperties(asDraft07), /*#__PURE__*/remapPropertyNames(asDraft07), /*#__PURE__*/remapItems(asDraft07), /*#__PURE__*/remapAnyOf(asDraft07), /*#__PURE__*/remapAllOf(asDraft07)); // ---------------------
// Json Schema 2020-12
// ---------------------

/** @internal */

export const as2020 = schema => pipe(schema, itemsToPrefixItems(as2020), recurse2020);
const itemsToPrefixItems = /*#__PURE__*/recurseKeyOf(JsonArray, 'items', () => schema => Array.isArray(schema) ? E.right(tuple('prefixItems', schema.map(as2020), {
  items: false
})) : E.left(tuple(as2020(schema))));
const recurse2020 = /*#__PURE__*/flow( /*#__PURE__*/remapDefs(as2020), /*#__PURE__*/remapContentSchema(as2020), /*#__PURE__*/remapProperties(as2020), /*#__PURE__*/remapAdditionalProperties(as2020), /*#__PURE__*/remapPropertyNames(as2020), /*#__PURE__*/remapAnyOf(as2020), /*#__PURE__*/remapAllOf(as2020));
//# sourceMappingURL=json-schema.mjs.map