/** @since 1.0.0 */
import { pipe } from "fp-ts/es6/function.js";
import * as k from 'kuvio';
import { deriveGuard } from 'schemata-ts/derivations/guard-schemable';
import { Brand } from 'schemata-ts/schemata/Brand';
import { Date as DateS } from 'schemata-ts/schemata/Date';
import { Imap } from 'schemata-ts/schemata/Imap';
import { Pattern } from 'schemata-ts/schemata/Pattern';
import { Refine } from 'schemata-ts/schemata/Refine';
/**
 * E.g. `2022`
 *
 * @since 1.0.0
 * @category Pattern
 */

const year = /*#__PURE__*/pipe(k.digit, /*#__PURE__*/k.exactly(4));
/**
 * E.g. `+-002022`
 *
 * @since 1.0.0
 * @category Pattern
 */

const expandedYear = /*#__PURE__*/pipe( /*#__PURE__*/k.characterClass(false, '+', '-'), k.subgroup, /*#__PURE__*/k.andThen( /*#__PURE__*/pipe(k.digit, /*#__PURE__*/k.exactly(6))));
/**
 * E.g. 01, 02, 03, 04, 05, 06, 07, 08, 09, 10, 11, 12
 *
 * @since 1.0.0
 * @category Pattern
 */

const month = /*#__PURE__*/pipe( /*#__PURE__*/k.char('0'), /*#__PURE__*/k.andThen( /*#__PURE__*/pipe( /*#__PURE__*/k.integerRange(1, 9), k.subgroup)), k.subgroup, /*#__PURE__*/k.or( /*#__PURE__*/pipe( /*#__PURE__*/k.integerRange(10, 12), k.subgroup)), k.subgroup);
/**
 * E.g. 01, 02, 03, 04, 05, ..., 30, 31
 *
 * @since 1.0.0
 * @category Pattern
 */

const day = /*#__PURE__*/pipe( /*#__PURE__*/k.char('0'), /*#__PURE__*/k.andThen( /*#__PURE__*/pipe( /*#__PURE__*/k.integerRange(1, 9), k.subgroup)), k.subgroup, /*#__PURE__*/k.or( /*#__PURE__*/pipe( /*#__PURE__*/k.integerRange(10, 31), k.subgroup)), k.subgroup);
/**
 * E.g. `YYYY-MM-DD`
 *
 * @since 1.0.0
 * @category Pattern
 */

const yearMonthDay = /*#__PURE__*/k.sequence( /*#__PURE__*/k.subgroup( /*#__PURE__*/k.oneOf(year, expandedYear)), /*#__PURE__*/k.char('-'), month, /*#__PURE__*/k.char('-'), day);
/**
 * E.g. `YYYY-MM`
 *
 * @since 1.0.0
 * @category Pattern
 */

const yearMonth = /*#__PURE__*/k.subgroup( /*#__PURE__*/k.sequence( /*#__PURE__*/k.subgroup( /*#__PURE__*/k.oneOf(year, expandedYear)), /*#__PURE__*/k.char('-'), month));
/**
 * E.G. `YYYY-MM-DD` or `YYYY-MM` or `YYYY`
 *
 * @since 1.0.0
 * @category Pattern
 */

const date = /*#__PURE__*/k.subgroup( /*#__PURE__*/k.oneOf(yearMonthDay, yearMonth, year, expandedYear));
/**
 * E.g. 00, 01, 02, 03, 04, 05, 06, 07, ..., 22, 23
 *
 * @since 1.0.0
 * @category Pattern
 */

const hour = /*#__PURE__*/pipe( /*#__PURE__*/k.char('0'), /*#__PURE__*/k.andThen( /*#__PURE__*/pipe( /*#__PURE__*/k.integerRange(0, 9), k.subgroup)), k.subgroup, /*#__PURE__*/k.or( /*#__PURE__*/pipe( /*#__PURE__*/k.integerRange(10, 23), k.subgroup)), k.subgroup);
/**
 * E.g. 00, 01, 02, 03, 04, 05, 06, 07, ..., 58, 59
 *
 * @since 1.0.0
 * @category Pattern
 */

export const minutesSeconds = /*#__PURE__*/pipe( /*#__PURE__*/k.char('0'), /*#__PURE__*/k.andThen( /*#__PURE__*/pipe( /*#__PURE__*/k.integerRange(0, 9), k.subgroup)), k.subgroup, /*#__PURE__*/k.or( /*#__PURE__*/pipe( /*#__PURE__*/k.integerRange(10, 59), k.subgroup)), k.subgroup);
/**
 * @since 1.0.0
 * @category Pattern
 */

const milliseconds = /*#__PURE__*/pipe(k.digit, /*#__PURE__*/k.atLeastOne(), k.subgroup);
/**
 * E.g. `Z` or `±05:00`
 *
 * @since 1.0.0
 * @category Pattern
 */

const timezoneOffset = /*#__PURE__*/pipe( /*#__PURE__*/k.char('Z'), /*#__PURE__*/k.or( /*#__PURE__*/pipe( /*#__PURE__*/k.characterClass(false, '+', '-'), /*#__PURE__*/k.andThen(hour), /*#__PURE__*/k.andThen( /*#__PURE__*/k.char(':')), /*#__PURE__*/k.andThen(minutesSeconds))), k.subgroup);
/**
 * E.g. `HH:mm:ss.SSS`
 *
 * @since 1.0.0
 * @category Pattern
 */

const hrMinSecMs = /*#__PURE__*/pipe(hour, /*#__PURE__*/k.andThen( /*#__PURE__*/k.char(':')), /*#__PURE__*/k.andThen(minutesSeconds), /*#__PURE__*/k.andThen( /*#__PURE__*/k.char(':')), /*#__PURE__*/k.andThen(minutesSeconds), /*#__PURE__*/k.andThen( /*#__PURE__*/k.char('.')), /*#__PURE__*/k.andThen(milliseconds));
/**
 * E.g. `HH:mm:ss`
 *
 * @since 1.0.0
 * @category Pattern
 */

const hrMinSec = /*#__PURE__*/pipe(hour, /*#__PURE__*/k.andThen( /*#__PURE__*/k.char(':')), /*#__PURE__*/k.andThen(minutesSeconds), /*#__PURE__*/k.andThen( /*#__PURE__*/k.char(':')), /*#__PURE__*/k.andThen(minutesSeconds));
/**
 * E.g. `HH:mm`
 *
 * @since 1.0.0
 * @category Pattern
 */

const hrMin = /*#__PURE__*/pipe(hour, /*#__PURE__*/k.andThen( /*#__PURE__*/k.char(':')), /*#__PURE__*/k.andThen(minutesSeconds));
/**
 * Iso time string
 *
 * @since 1.0.0
 * @category Pattern
 */

const time = /*#__PURE__*/pipe(hrMinSecMs, /*#__PURE__*/k.or(hrMinSec), /*#__PURE__*/k.or(hrMin), k.subgroup);
/**
 * @since 1.0.0
 * @category Pattern
 */

const dateTimeStringOptT = /*#__PURE__*/pipe(date, /*#__PURE__*/k.andThen( /*#__PURE__*/pipe( /*#__PURE__*/k.char('T'), /*#__PURE__*/k.or( /*#__PURE__*/k.char(' ')), k.subgroup, /*#__PURE__*/k.andThen(time), k.subgroup, k.maybe)));
/**
 * @since 1.0.0
 * @category Pattern
 */

const dateTimeStringReqT = /*#__PURE__*/pipe(date, /*#__PURE__*/k.andThen( /*#__PURE__*/pipe( /*#__PURE__*/k.char('T'), /*#__PURE__*/k.or( /*#__PURE__*/k.char(' ')), k.subgroup)), /*#__PURE__*/k.andThen(time));
/**
 * @since 1.0.0
 * @category Pattern
 */

const isoDateStringOptTzOptT = /*#__PURE__*/pipe(dateTimeStringOptT, /*#__PURE__*/k.andThen( /*#__PURE__*/pipe(timezoneOffset, k.maybe)));
/**
 * @since 1.0.0
 * @category Pattern
 */

const isoDateStringOptTzReqT = /*#__PURE__*/pipe(dateTimeStringReqT, /*#__PURE__*/k.andThen( /*#__PURE__*/pipe(timezoneOffset, k.maybe)));
/**
 * @since 1.0.0
 * @category Pattern
 */

const isoDateStringReqTzReqT = /*#__PURE__*/pipe(dateTimeStringReqT, /*#__PURE__*/k.andThen(timezoneOffset));
/**
 * The Date parser (used in DateFromString) accepts different strings depending on
 * runtime, and also accepts other formats like `February 29, 2022`.
 *
 * `DateFromIsoString` follows a subset of the [ECMAScript 2023 Language Date Time String
 * Specification](https://tc39.es/ecma262/#sec-date-time-string-format).
 *
 * Notable features:
 *
 * - Requires padded months, days, hours, minutes, and seconds
 * - Can be configured to require a time, time and timezone offset (e.g. `Z` or `±05:00`) or
 *   neither (default is require both).
 * - Dates may contain years, months, and days; years and months; or years
 * - Times may contain hours, minutes, seconds, and milliseconds; hours, minutes, and
 *   seconds; or hours and minutes.
 * - Expanded years are permitted (e.g. `+002022` instead of `2022`)
 *
 * @since 1.0.0
 * @category Conversion
 */

export const DateFromIsoString = (params = {}) => {
  const {
    requireTime = 'TimeAndOffset'
  } = params;

  const pattern = (() => {
    switch (requireTime) {
      case 'None':
        return isoDateStringOptTzOptT;

      case 'Time':
        return isoDateStringOptTzReqT;

      case 'TimeAndOffset':
        return isoDateStringReqTzReqT;
    }
  })();

  return pipe(Pattern(pattern, requireTime === 'TimeAndOffset' ? 'IsoDateTimeStringZ' : requireTime === 'Time' ? 'IsoDateTimeString' : 'IsoDateString'), Refine(s => !Number.isNaN(new Date(s).getTime()), 'SafeDate'), Brand(), Imap(deriveGuard(DateS()), s => new Date(s), d => d.toISOString()));
};
//# sourceMappingURL=DateFromIsoString.mjs.map