/** @since 1.0.0 */
import { unsafeCoerce } from "fp-ts/es6/function.js";
import { SchemaImplementation } from 'schemata-ts/Schema';
/**
 * Floating point branded newtype. Parameters: min, max are inclusive.
 *
 * Represents floating point numbers:
 *
 * ```math
 *  { f | f ∈ ℝ, f >= -Number.MAX_VALUE, f <= Number.MAX_VALUE }
 * ```
 *
 * @since 1.0.0
 * @category Number
 */

export const Float = params => new FloatSchema(params);
/**
 * The FloatSchema transformer class, use instead `S.Float` function to create a FloatSchema
 *
 * @since 2.2.0
 * @category Transformations
 */

export class FloatSchema extends SchemaImplementation {
  params;

  constructor(params) {
    super(s => s.float(params));
    this.params = params;
  }
  /**
   * Brands this FloatSchema with a certain brand
   *
   * @since 2.2.0
   */


  brand = () => unsafeCoerce(this);
  /**
   * Sets the minimum value of the FloatSchema
   *
   * @since 2.2.0
   */

  min = min => new FloatSchema({ ...this.params,
    min
  });
  /**
   * Sets the maximum value of the FloatSchema
   *
   * @since 2.2.0
   */

  max = max => new FloatSchema({ ...this.params,
    max
  });
  /**
   * Overrides the 'expected' field in TranscodeError > TypeMismatch
   *
   * @since 2.2.0
   * @default 'string<minLength?, maxLength?>'
   */

  errorName = errorName => new FloatSchema({ ...this.params,
    errorName
  });
}
//# sourceMappingURL=Float.mjs.map