/** @since 1.0.0 */
import { pipe } from "fp-ts/es6/function.js";
import * as k from 'kuvio';
import { PrimitivesGuard } from 'schemata-ts/schemables/primitives/instances/guard';
import { Brand } from 'schemata-ts/schemata/Brand';
import { Imap } from 'schemata-ts/schemata/Imap';
import { Pattern } from 'schemata-ts/schemata/Pattern';
import { Refine } from 'schemata-ts/schemata/Refine';
/**
 * Negative floats with at least one digit before the decimal point.
 *
 * E.g. `-1.`; but neither `-.` nor `-`
 *
 * @since 1.0.0
 * @category Pattern
 */

const negativeFloatLeft = /*#__PURE__*/pipe( /*#__PURE__*/k.char('-'), /*#__PURE__*/k.andThen( /*#__PURE__*/pipe(k.digit, /*#__PURE__*/k.between(1, 308))), /*#__PURE__*/k.andThen( /*#__PURE__*/pipe( /*#__PURE__*/k.char('.'), /*#__PURE__*/k.andThen( /*#__PURE__*/pipe(k.digit, /*#__PURE__*/k.anyNumber())), k.subgroup, k.maybe)));
/**
 * Negative floats with at least one digit after the decimal point.
 *
 * E.g. `-.1`; but neither `-.` nor `-`
 *
 * @since 1.0.0
 * @category Pattern
 */

const negativeFloatRight = /*#__PURE__*/pipe( /*#__PURE__*/k.char('-'), /*#__PURE__*/k.andThen( /*#__PURE__*/pipe(k.digit, /*#__PURE__*/k.between(0, 308))), /*#__PURE__*/k.andThen( /*#__PURE__*/pipe( /*#__PURE__*/k.char('.'), /*#__PURE__*/k.andThen( /*#__PURE__*/pipe(k.digit, /*#__PURE__*/k.atLeastOne())), k.subgroup, k.maybe)));
/**
 * Positive floats with at least one digit before the decimal point.
 *
 * E.g. `1.` but not `.`
 *
 * @since 1.0.0
 * @category Pattern
 */

const positiveFloatLeft = /*#__PURE__*/pipe(k.digit, /*#__PURE__*/k.between(1, 308), /*#__PURE__*/k.andThen( /*#__PURE__*/pipe( /*#__PURE__*/k.char('.'), /*#__PURE__*/k.andThen( /*#__PURE__*/pipe(k.digit, /*#__PURE__*/k.anyNumber())), k.subgroup, k.maybe)));
/**
 * Positive floats with at least one digit after the decimal point.
 *
 * E.g. `.1` but not `.`
 *
 * @since 1.0.0
 * @category Pattern
 */

const positiveFloatRight = /*#__PURE__*/pipe(k.digit, /*#__PURE__*/k.between(0, 308), /*#__PURE__*/k.andThen( /*#__PURE__*/pipe( /*#__PURE__*/k.char('.'), /*#__PURE__*/k.andThen( /*#__PURE__*/pipe(k.digit, /*#__PURE__*/k.atLeastOne())), k.subgroup, k.maybe)));
/**
 * @since 1.0.0
 * @category Pattern
 */

const decimalFloat = /*#__PURE__*/k.oneOf(negativeFloatLeft, negativeFloatRight, positiveFloatLeft, positiveFloatRight);
/**
 * @since 1.0.0
 * @category Pattern
 */

const positiveExponential = /*#__PURE__*/pipe( /*#__PURE__*/k.char('+'), k.maybe, /*#__PURE__*/k.andThen( /*#__PURE__*/pipe( /*#__PURE__*/k.integerRange(0, 308), k.subgroup)));
/**
 * @since 1.0.0
 * @category Pattern
 */

const negativeExponential = /*#__PURE__*/pipe( /*#__PURE__*/k.char('-'), /*#__PURE__*/k.andThen( /*#__PURE__*/pipe(k.digit, /*#__PURE__*/k.atLeastOne(), k.subgroup)));
/**
 * @since 1.0.0
 * @category Pattern
 */

const floatFromString = /*#__PURE__*/pipe(decimalFloat, k.subgroup, /*#__PURE__*/k.andThen( /*#__PURE__*/pipe( /*#__PURE__*/k.char('e'), /*#__PURE__*/k.andThen( /*#__PURE__*/pipe( /*#__PURE__*/k.oneOf(positiveExponential, negativeExponential), k.subgroup)), k.subgroup, k.maybe)));
/**
 * Floating point branded newtype from strings. Parameters: min, max are inclusive.
 *
 * Represents string floating point numbers:
 *
 * ```math
 *  { f | f ∈ ℝ, f >= -Number.MAX_VALUE, f <= Number.MAX_VALUE }
 * ```
 *
 * @deprecated Use `ParseFloat` instead.
 * @since 1.0.0
 * @category Conversion
 */

export const FloatFromString = /*#__PURE__*/pipe( /*#__PURE__*/Pattern(floatFromString, ['FloatString', 'Float']), /*#__PURE__*/Refine(s => PrimitivesGuard.float().is(Number(s)) && s.trim() !== '', 'Float'), /*#__PURE__*/Brand(), /*#__PURE__*/Imap( /*#__PURE__*/PrimitivesGuard.float(), s => Number(s), n => n.toString()));
//# sourceMappingURL=FloatFromString.mjs.map