/** @since 1.4.0 */
import { pipe, unsafeCoerce } from "fp-ts/es6/function.js";
import * as O from "fp-ts/es6/Option.js";
import * as RR from "fp-ts/es6/ReadonlyRecord.js";
import * as Sg from "fp-ts/es6/Semigroup.js";
import { deriveGuard } from 'schemata-ts/derivations/guard-schemable';
import { deriveInformation } from 'schemata-ts/derivations/information-schemable';
import { deriveTypeString } from 'schemata-ts/derivations/type-string-schemable';
import { hasOwn } from 'schemata-ts/internal/util';
import { SchemaImplementation } from 'schemata-ts/Schema';
import { StructTypeString } from 'schemata-ts/schemables/struct/instances/type-string';
import { Imap } from 'schemata-ts/schemata/Imap';
import { Option } from 'schemata-ts/schemata/Option';
import { Optional } from 'schemata-ts/schemata/Optional';
import { Readonly } from 'schemata-ts/schemata/Readonly';
/**
 * Used to construct a struct schema with enumerated keys.
 *
 * **Note:** Index signatures must accomodate the input/output types for all other
 * enumerated keys. It will decode properly otherwise, but TypeScript will not permit
 * construction of such a type
 *
 * **Note:** The second parameter `extraProps` is deprecated, use `Struct({}).strict()` or
 * `Struct({}).addIndexSignature()` instead
 *
 * @since 1.0.0
 * @category Combinators
 * @see https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures
 */

export const Struct = (props, extraProps = 'strip') => new StructSchema(props, extraProps);
/**
 * Use `Struct({})` schema combinator instead
 *
 * @since 2.1.0
 * @category Transformations
 */

export class StructSchema extends SchemaImplementation {
  definition;
  indexSignature;

  constructor(definition, indexSignature) {
    super(_ => {
      const struct = {};
      const structName = {};

      for (const key in definition) {
        if (!hasOwn(definition, key)) continue; // eslint-disable-next-line @typescript-eslint/no-non-null-assertion

        const schema = definition[key];
        const schemable = schema.runSchema(_);
        const guard = deriveGuard(schema);
        const information = deriveInformation(schema);
        const name = deriveTypeString(schema);
        struct[key] = {
          schemable,
          guard,
          information,
          semigroup: Sg.last(),
          name: name[0]
        };
        structName[key] = {
          schemable: name,
          guard,
          information,
          semigroup: Sg.last(),
          name: name[0]
        };
      }

      const extraProps_ = indexSignature === undefined || typeof indexSignature === 'string' ? indexSignature : indexSignature.runSchema(_);
      const extraPropsName_ = indexSignature === undefined || typeof indexSignature === 'string' ? indexSignature : deriveTypeString(indexSignature);
      const wholeName = StructTypeString.struct(structName, extraPropsName_, '');
      return _.struct(struct, extraProps_, wholeName[0]);
    });
    this.definition = definition;
    this.indexSignature = indexSignature;
  }
  /**
   * Re-declares a StructSchema by only including specified properties.
   *
   * Returns a new `StructSchema`
   *
   * @since 2.1.0
   */


  pick = (...keys) => new StructSchema(pipe(this.definition, RR.filterWithIndex(k => keys.includes(k)), _ => unsafeCoerce(_)), this.indexSignature);
  /**
   * Re-declares a StructSchema by excluding specified properties.
   *
   * Returns a new `StructSchema`
   *
   * @since 2.1.0
   */

  omit = (...keys) => new StructSchema(pipe(this.definition, RR.filterWithIndex(k => !keys.includes(k)), _ => unsafeCoerce(_)), this.indexSignature);
  /**
   * Marks all properties as optional; applies `Partial` to both input and output types.
   *
   * Returns a new schema.
   *
   * @since 2.1.0
   */

  partial = () => unsafeCoerce(new StructSchema(pipe(this.definition, RR.map(Optional)), this.indexSignature));
  /**
   * A variant of `partial` that applies `Partial` to input properties and maps each
   * output property to the fp-ts `Option` type.
   *
   * Returns a new schema.
   *
   * @since 2.1.0
   */

  partialOption = () => unsafeCoerce(new StructSchema(pipe(this.definition, RR.map(OptionFromOptional)), this.indexSignature));
  /**
   * Marks all properties as readonly; applies `Readonly` to both input and output types.
   *
   * Returns a new schema.
   *
   * @since 2.1.0
   */

  readonly = () => Readonly(new StructSchema(this.definition, this.indexSignature));
  /**
   * Sets a Struct Schema's index signature to be strict
   *
   * Returns a new `StructSchema`.
   *
   * @since 2.1.0
   */

  strict = () => new StructSchema(this.definition, 'error');
  /**
   * Adds an index signature to a Struct Schema.
   *
   * Returns a new `StructSchema`.
   *
   * @since 2.1.0
   */

  addIndexSignature = indexSignature => new StructSchema(this.definition, indexSignature);
  /**
   * Extends a Struct Schema with additional properties. Keys specified in `props` will
   * overwrite keys in `this`.
   *
   * Returns a new `StructSchema`.
   *
   * @since 2.1.0
   */

  extend = props => this.intersect(new StructSchema(props, this.indexSignature));
  /**
   * Intersects the present Struct Schema with another effectively concatenating their
   * keys. Keys in `this` will be overwritten with identical keys in `that`.
   *
   * **Note:** The index signature of `that` will be discarded.
   *
   * Returns a new `StructSchema`.
   *
   * @since 2.1.0
   */

  intersect = that => new StructSchema({ ...this.definition,
    ...that.definition
  }, this.indexSignature);
}

const OptionFromOptional = schema => {
  const guard = deriveGuard(Option(schema));
  return pipe(Optional(schema), Imap({
    is: u => guard.is(u) && pipe(u, O.fold(() => true, a => a !== undefined))
  }, O.fromNullable, O.toUndefined));
};
//# sourceMappingURL=Struct.mjs.map