import { type OptionalInputProps, type OptionOutputProps, type OutputProps, type RequiredInputProps, type RestInput, type RestOutput } from 'schemata-ts/internal/schema';
import { type Schema, SchemaImplementation } from 'schemata-ts/Schema';
import { Readonly } from 'schemata-ts/schemata/Readonly';
import { type Simplify, type Spread } from 'type-fest';
/**
 * Used to construct a struct schema with enumerated keys.
 *
 * **Note:** Index signatures must accomodate the input/output types for all other
 * enumerated keys. It will decode properly otherwise, but TypeScript will not permit
 * construction of such a type
 *
 * **Note:** The second parameter `extraProps` is deprecated, use `Struct({}).strict()` or
 * `Struct({}).addIndexSignature()` instead
 *
 * @since 1.0.0
 * @category Combinators
 * @see https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures
 */
export declare const Struct: <T extends PropBase, Ix extends IxSigBase = undefined>(props: T, extraProps?: ExtraProps<Ix>) => StructSchema<T, Ix>;
type IxSigBase = Schema<any, any> | undefined;
type PropBase = Record<string, Schema<any, any>>;
type ExtraProps<Ix extends IxSigBase> = 'strip' | 'error' | Ix;
type Input<T extends PropBase, Ix extends IxSigBase> = Simplify<RestInput<Ix> & RequiredInputProps<T> & OptionalInputProps<T>>;
type Output<T extends PropBase, Ix extends IxSigBase> = Simplify<RestOutput<Ix> & OutputProps<T>>;
type OptionOutput<T extends PropBase, Ix extends IxSigBase> = Simplify<RestOutput<Ix> & OptionOutputProps<T>>;
/**
 * Use `Struct({})` schema combinator instead
 *
 * @since 2.1.0
 * @category Transformations
 */
export declare class StructSchema<T extends PropBase, Ix extends IxSigBase> extends SchemaImplementation<Input<T, Ix>, Output<T, Ix>> implements Schema<Input<T, Ix>, Output<T, Ix>> {
    readonly definition: T;
    private readonly indexSignature?;
    constructor(definition: T, indexSignature?: ExtraProps<Ix> | undefined);
    /**
     * Re-declares a StructSchema by only including specified properties.
     *
     * Returns a new `StructSchema`
     *
     * @since 2.1.0
     */
    readonly pick: <K extends keyof T>(...keys: readonly K[]) => StructSchema<Pick<T, K> extends infer T_1 ? { [KeyType_1 in keyof T_1]: Pick<T, K>[KeyType_1]; } : never, Ix>;
    /**
     * Re-declares a StructSchema by excluding specified properties.
     *
     * Returns a new `StructSchema`
     *
     * @since 2.1.0
     */
    readonly omit: <K extends keyof T>(...keys: readonly K[]) => StructSchema<Omit<T, K> extends infer T_1 ? { [KeyType_1 in keyof T_1]: Omit<T, K>[KeyType_1]; } : never, Ix>;
    /**
     * Marks all properties as optional; applies `Partial` to both input and output types.
     *
     * Returns a new schema.
     *
     * @since 2.1.0
     */
    readonly partial: () => Schema<Simplify<Partial<Input<T, Ix>>>, Simplify<Partial<Output<T, Ix>>>>;
    /**
     * A variant of `partial` that applies `Partial` to input properties and maps each
     * output property to the fp-ts `Option` type.
     *
     * Returns a new schema.
     *
     * @since 2.1.0
     */
    readonly partialOption: () => Schema<Simplify<Partial<Input<T, Ix>>>, Simplify<OptionOutput<T, Ix>>>;
    /**
     * Marks all properties as readonly; applies `Readonly` to both input and output types.
     *
     * Returns a new schema.
     *
     * @since 2.1.0
     */
    readonly readonly: () => Schema<Simplify<Readonly<Input<T, Ix>>>, Simplify<Readonly<Output<T, Ix>>>>;
    /**
     * Sets a Struct Schema's index signature to be strict
     *
     * Returns a new `StructSchema`.
     *
     * @since 2.1.0
     */
    readonly strict: () => StructSchema<T, undefined>;
    /**
     * Adds an index signature to a Struct Schema.
     *
     * Returns a new `StructSchema`.
     *
     * @since 2.1.0
     */
    readonly addIndexSignature: <Ix2 extends Schema<any, any>>(indexSignature: Ix2) => StructSchema<T, Ix2>;
    /**
     * Extends a Struct Schema with additional properties. Keys specified in `props` will
     * overwrite keys in `this`.
     *
     * Returns a new `StructSchema`.
     *
     * @since 2.1.0
     */
    readonly extend: <T2 extends PropBase>(props: T2) => StructSchema<Spread<T, T2>, Ix>;
    /**
     * Intersects the present Struct Schema with another effectively concatenating their
     * keys. Keys in `this` will be overwritten with identical keys in `that`.
     *
     * **Note:** The index signature of `that` will be discarded.
     *
     * Returns a new `StructSchema`.
     *
     * @since 2.1.0
     */
    readonly intersect: <T2 extends PropBase>(that: StructSchema<T2, any>) => StructSchema<Spread<T, T2>, Ix>;
}
export {};
//# sourceMappingURL=Struct.d.ts.map